/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#ifdef MALLOC_H
#include <malloc.h>
#endif
#include "is_types.h"
#include "isam.h"

#ifndef lint
static	char	sccsid[] = "@(#)db_create.c	1.8	10:32:27	2/26/92";
#endif

extern	dbm_t	*db_open();
extern	file_t	*isam_file_open();
#ifndef	MALLOC_H
extern	char	*malloc();
extern	char	*realloc();
#endif

#ifdef	BSD
#define	memcpy(d,s,n)	bcopy((s),(d),(n))
#else
#define	bzero(d,n)	memset((d), 0, (n))
#endif

extern	int	db_errno;

/*
 * db_create - create a new data dictionary
 *
 * db_create() creates the ISAM data dictionary file for `name'.  It will
 * be initialized to contain no file or field entries.
 */

dbm_t *
db_create (name)
char	*name;
{
	dbm_t	*dbm;

	/*
	 * db_open knows how to create dictionaries, use it to do
	 * all the work.
	 */

	if ((dbm = db_open (name, O_RDWR|O_CREAT|O_EXCL)) == 0)
		return 0;

	return dbm;
}

/*
 * db_create_file - create a new database file and update the dictionary
 *
 * db_create_file() creates the ISAM files needed for a database file and
 * updates the data dictionary.
 */

int
db_create_file (dbm, name)
dbm_t	*dbm;
char	*name;
{
	int	i;
	int	hole = -1;		/* first free space in file table */
	file_t	*file;
	file_t	*files;
	field_t	*field;
	is_index_t *index;
	sort_key_t *key;
	schema_t *schema;
	char	short_name[MAXFILENAME+1];
	struct	{
		int	f_fileno;
		int	f_fieldno;
		int	f_fieldtype;
		int	f_fieldsize;
		char	f_fieldfile[MAXFILENAME];
		char	f_fieldname[MAXFIELDNAME];
		short	f_fieldkeys[MAXSORTFIELDS];
	} record;

	/*
	 * Enforce the file name length by truncating it to the shorter
	 * length.  
	 */

	if (strlen (name) > MAXFILENAME) {
		strncpy (short_name, name, MAXFILENAME);
		short_name[MAXFILENAME] = '\0';
		name = short_name;
	}

	/*
	 * Check to see if this is an already existing file.  We also
	 * see if there are any unused file numbers that can be taken.
	 */

	if (dbm->d_files) {
		for (i = 0;i < dbm->d_file_count;i++) {
			if (dbm->d_files[i].f_file_name[0] == '\0') {
				if (hole == -1)
					hole = i;

				continue;
			}
			if (! strncmp (dbm->d_files[i].f_file_name,
					name, MAXFILENAME))
				return (db_errno = DB_FILE_EXISTS), -1;
		}
	}
	if ((dbm->d_mode & (O_RDONLY|O_WRONLY|O_RDWR)) != O_RDWR)
		return (db_errno = DB_INVALID), -1;

	if (! (file = isam_file_open (name,
			dbm->d_mode | (O_CREAT|O_EXCL))))
		return -1;

	if (! (file->f_schema = (schema_t *) malloc (sizeof (schema_t))))
		return (db_errno = DB_NOMEM), -1;

	file->f_schema->s_field_count = 0;
	file->f_schema->s_fields = 0;

	if (dbm->d_files) {
		if (hole == -1) {
			if (! (files = (file_t *) realloc (dbm->d_files,
				sizeof *file * (dbm->d_file_count + 1))))
				return (db_errno = DB_NOMEM), -1;

			hole = dbm->d_file_count;

			dbm->d_files = files;
			dbm->d_file_count++;
		}
	} else {
		if (! (files = (file_t *) malloc (sizeof *file)))
			return (db_errno = DB_NOMEM), -1;

		hole = 0;

		dbm->d_files = files;
		dbm->d_file_count = 1;
	}
	dbm->d_files[hole] = *file;
	file = &dbm->d_files[hole];

	schema = file->f_schema;
	schema->s_field_count = 2;
	if (! (schema->s_fields = (field_t *) malloc (2 * sizeof (field_t))))
		return (db_errno = DB_NOMEM), -1;

	field = &schema->s_fields[1];
	field->f_type = T_INDEX;
	field->f_size = 0;
	field->f_offset = 1;
	field->f_name[0] = '.';
	strncpy (field->f_name + 1, name, (MAXFIELDNAME-1));

	if (! (index = (is_index_t *) malloc (sizeof *index)))
		return (db_errno = DB_NOMEM), -1;

	file->f_index = index;
	index->i_key_count = 1;
	if (! (index->i_keys = (sort_key_t *) malloc (2 * sizeof (sort_key_t))))
		return (db_errno = DB_NOMEM), -1;

	key = &index->i_keys[1];
	key->k_field_number = 1;
	key->k_field_count = 1;
	key->k_key_size = sizeof (off_t);
	key->k_unique = 0;
	key->k_fields[0] = 0;

	bzero (&record, sizeof record);
	record.f_fileno = hole;
	record.f_fieldno = 1;
	record.f_fieldtype = field->f_type;
	record.f_fieldsize = field->f_size;
	strncpy (record.f_fieldfile, file->f_file_name, MAXFILENAME);
	strncpy (record.f_fieldname, field->f_name, MAXFIELDNAME);

	if (__db_write_record (&dbm->d_dictionary, &record))
		return -1;

	return hole;
}

int
db_create_field (dbm, filenum, field)
dbm_t	*dbm;
int	filenum;
field_t	*field;
{
	file_t	*file;
	field_t	*fields;
	schema_t *schema;
	int	i;
	int	fieldnum;
	struct	{
		int	f_fileno;
		int	f_fieldno;
		int	f_fieldtype;
		int	f_fieldsize;
		char	f_fieldfile[MAXFILENAME];
		char	f_fieldname[MAXFIELDNAME];
		short	f_fieldkeys[MAXSORTFIELDS];
	} record;
	struct	stat	stat;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (fstat (file->f_data_fd, &stat))
		return (db_errno = DB_EIO), -1;

	if (stat.st_size > sizeof (off_t))
		return (db_errno = DB_NOT_EMPTY), -1;

	schema = file->f_schema;

	if (schema->s_fields) {
		fields = schema->s_fields;

		for (i = 1;i < schema->s_field_count;i++) {
			if (strncmp (field->f_name, fields[i].f_name,
					MAXFIELDNAME) == 0)
				return (db_errno = DB_FIELD_EXISTS), -1;
		}
	} else
		fields = 0;

	if (fields && field->f_type == T_SERIAL) {
		for (i = 1;i < schema->s_field_count;i++)
			if (fields[i].f_type == T_SERIAL)
				return (db_errno = DB_INVALID), -1;
	}
	if (field->f_type != T_CHAR) {
		switch (field->f_type) {
			case T_SHORT:
				field->f_size = sizeof (short);
				break;
			case T_INT:
				field->f_size = sizeof (int);
				break;
			case T_LONG:
			case T_DATE:
			case T_SERIAL:
				field->f_size = sizeof (long);
				break;
			case T_FLOAT:
				field->f_size = sizeof (float);
				break;
			case T_DOUBLE:
			case T_MONEY:
				field->f_size = sizeof (double);
				break;
			default:
				return (db_errno = DB_INVALID), -1;
		}
	} else if (field->f_type == T_CHAR) {
		if (field->f_size <= 0 || field->f_size > 256)
			return (db_errno = DB_INVALID), -1;
	}
	if (fields) {
		if (! (fields = (field_t *) realloc (fields,
				(schema->s_field_count + 1) * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		fieldnum = schema->s_field_count++;
		bzero (&fields[fieldnum], sizeof *fields);
		schema->s_fields = fields;
	} else {
		fieldnum = 1;
		if (! (fields = (field_t *) malloc (2 * sizeof (field_t))))
			return (db_errno = DB_NOMEM), -1;

		bzero (fields, 2 * sizeof (field_t));
		schema->s_field_count = 2;
		schema->s_fields = fields;
	}
	bzero (&record, sizeof record);
	record.f_fileno = filenum;
	record.f_fieldno = fieldnum;
	record.f_fieldtype = field->f_type;
	record.f_fieldsize = field->f_size;
	strncpy (record.f_fieldfile, file->f_file_name, MAXFILENAME);
	strncpy (record.f_fieldname, field->f_name, MAXFIELDNAME);

	if (__db_write_record (&dbm->d_dictionary, &record))
		goto fail;

	schema->s_fields[fieldnum] = *field;
	file->f_record_size += field->f_size;
	return fieldnum;

fail:
	if (schema->s_field_count == 2) {
		free (schema->s_fields);
		schema->s_fields = 0;
		schema->s_field_count = 0;
	} else
		schema->s_field_count--;

	return -1;
}

/*
 * db_create_index - create a new database index
 *
 * db_create_index() creates a new ISAM index.  The named file must be
 * empty and the list of fields must exist.
 */

int
db_create_index (dbm, filenum, field, key)
dbm_t	*dbm;
int	filenum;
field_t	*field;
sort_key_t *key;
{
	file_t	*file;
	field_t	*fields;
	field_t *member;
	sort_key_t *keys;
	schema_t *schema;
	is_index_t *index;
	int	i, j;
	int	fieldnum;
	struct	{
		int	f_fileno;
		int	f_fieldno;
		int	f_fieldtype;
		int	f_fieldsize;
		char	f_fieldfile[MAXFILENAME];
		char	f_fieldname[MAXFIELDNAME];
		short	f_fieldkeys[MAXSORTFIELDS];
	} record;
	struct	stat	stat;

	/*
	 * Validate the file number
	 */

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	if (field->f_type != T_INDEX)
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (fstat (file->f_data_fd, &stat))
		return (db_errno = DB_EIO), -1;

	if (stat.st_size > sizeof (off_t))
		return (db_errno = DB_NOT_EMPTY), -1;

	schema = file->f_schema;
	index = file->f_index;

	if (! schema->s_fields)
		return (db_errno = DB_INVALID), -1;

	fields = schema->s_fields;

	/*
	 * Make certain this index name is not already in use.
	 */

	for (i = 1;i < schema->s_field_count;i++) {
		if (strncmp (field->f_name, fields[i].f_name,
				MAXFIELDNAME) == 0)
			return (db_errno = DB_FIELD_EXISTS), -1;
	}
	if (key->k_field_count == 0 || key->k_field_count > MAXSORTFIELDS)
		return (db_errno = DB_INVALID), -1;

	if (key->k_unique)
		field->f_size = key->k_field_count;
	else
		field->f_size = - key->k_field_count;

	for (key->k_key_size = i = 0;i < key->k_field_count;i++) {
		j = key->k_fields[i] & ~IS_DESCENDING;
		if (j == 0)
			return (db_errno = DB_INVALID), -1;

		if (j >= schema->s_field_count)
			return (db_errno = DB_INVALID), -1;

		member = &fields[j];
		if (member->f_type == T_INDEX)
			return (db_errno = DB_INVALID), -1;

		key->k_key_size += member->f_size;
	}
	fieldnum = schema->s_field_count++;
	if (! (fields = (field_t *) realloc (fields,
			schema->s_field_count * sizeof (field_t)))) {
		schema->s_field_count--;
		return (db_errno = DB_NOMEM), -1;
	}
	schema->s_fields = fields;
	schema->s_fields[fieldnum] = *field;

	if (index) {
		index->i_key_count++;
		if (! (keys = (sort_key_t *) realloc (index->i_keys,
				sizeof *keys * index->i_key_count))) {
			db_errno = DB_NOMEM;
			goto fail;
		}
		index->i_keys = keys;
	} else {
		if (! (index = (is_index_t *) malloc (sizeof *index))) {
			index->i_key_count--;
			schema->s_field_count--;
			db_errno = DB_NOMEM;
			return -1;
		}
		file->f_index = index;
		index->i_key_count = 2;

		if (! (keys = (sort_key_t *) malloc (2 * sizeof *keys))) {
			db_errno = DB_NOMEM;
			goto fail;
		}
		index->i_keys = keys;
	}
	fields[fieldnum].f_offset = index->i_key_count - 1;

	key->k_field_number = schema->s_field_count;
	bzero (&record, sizeof record);
	record.f_fileno = filenum;
	record.f_fieldno = fieldnum;
	record.f_fieldtype = T_INDEX;

	if (key->k_unique)
		record.f_fieldsize = key->k_field_count;
	else
		record.f_fieldsize = - key->k_field_count;

	strncpy (record.f_fieldfile, file->f_file_name, MAXFILENAME);
	strncpy (record.f_fieldname, field->f_name, MAXFIELDNAME);
	memcpy (record.f_fieldkeys, key->k_fields, sizeof record.f_fieldkeys);

	if (__db_write_record (&dbm->d_dictionary, &record))
		goto fail;

	if (key->k_unique == 0) {
		key->k_field_count++;
		key->k_key_size += sizeof (off_t);
	}
	index->i_keys[index->i_key_count - 1] = *key;
	return index->i_key_count - 1;

fail:
	if (index->i_key_count == 2) {
		free (file->f_index);
		file->f_index = 0;
	} else
		index->i_key_count--;

	schema->s_field_count--;
	return -1;
}
