/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include "is_types.h"

#ifndef lint
static	char	sccsid[] = "@(#)db_del.c	1.9	11:50:36	1/29/92";
#endif

extern	int	isam_delete();
extern	int	db_errno;

int
__db_delete_key (file, index, data, record)
file_t	*file;
int	index;
VOIDP	*data;
off_t	record;
{
	char	buf[MAXBLOCKSIZE];

	__isam_buf2key (file, index, data, record, buf);
	return isam_delete (file, index, buf, record);
}

int
__db_delete_record (file)
file_t	*file;
{
	int	i;
	int	size = file->f_record_size;
	off_t	record = file->f_current_record;
	off_t	next;
	field_t	*fields;
	struct	stat	stat;
	char	*data = file->f_data_buffer;
	int	was_locked;
	long	ticket;

	/*
	 * There must be a current record to delete.  This record can
	 * be found by any function that sets the current record value.
	 */

	if (record <= 0)
		return (db_errno = DB_NO_RECORD), -1;

	/*
	 * The file must be locked the entire time the record is being
	 * deleted so that all keys are removed at the same time.
	 */

	if (! (was_locked = file->f_locked))
		if ((ticket = isam_set_ticket (file)) == -1)
			return -1;

	/*
	 * The current record is read in so that the keys can be
	 * computed.
	 */

	if (isam_read_data (file, record, data))
		goto fail;

	/*
	 * Delete the keys for all of the indexes.
	 */

	for (i = 1;i <= file->f_index->i_key_count;i++)
		if (__db_delete_key (file, i, data, record))
			goto fail;

	/*
	 * Return the data record to the free data record list.
	 */

	if (__db_free (file, record))
		goto fail;

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return 0;

fail:
	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return -1;
}

int
db_delete_record (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	file_t	*file;
	int	rc;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	rc = __db_delete_record (file);
	file->f_current_record = -1;

	return rc;
}

/*
 * db_delete_file - remove a file from the database
 *
 * db_delete_file() deletes a file and all associated information from the
 * data dictionary.
 */

int
db_delete_file (dbm, file)
dbm_t	*dbm;
char	*file;
{
	file_t	*dict = &dbm->d_dictionary;
	dict_t	record;
	int	i;
	int	was_locked = dict->f_locked;
	long	ticket;

	/*
	 * Locate the file in the dictionary.
	 */

	for (i = 0;i < dbm->d_file_count;i++)
		if (strcmp (file, dbm->d_files[i].f_file_name) == 0)
			break;

	if (i == dbm->d_file_count || dbm->d_files[i].f_file_name[0] == '\0')
		return (db_errno = DB_NO_FILE), -1;

	/*
	 * We lock the dictionary the entire time
	 */

	if (! was_locked) {
		if ((ticket = isam_set_ticket (dict)) == -1)
			return -1;
	}

	/*
	 * We know the file number (i), so we can loop through the file
	 * number index (index 1) searching for exact matches.  When
	 * we no longer are able to read records, we stop deleting them.
	 */

	while (! __db_read_record (dict, 1, (void *) &i, IS_EQUAL, &record)) {
		if (strncmp (record.f_fieldfile, file, MAXFILENAME) != 0) {
			db_errno = DB_BAD_DB;
			goto fail;
		}
		if (__db_delete_record (dict))
			goto fail;
	}

	/*
	 * The physical file now needs to be removed.  We check to see
	 * if it needs to be closed first, then we ask the low level code
	 * to do the dirty work.
	 */

	if (dbm->d_files[i].f_schema->s_fields) {
		close (dbm->d_files[i].f_data_fd);
		close (dbm->d_files[i].f_key_fd);
	}
	if (isam_delete_file (file))
		goto fail;

	if (! was_locked) {
		dict->f_ticket = ticket;
		isam_end_ticket (dict);
	}
	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (dict);

	return -1;
}
