/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#include "is_types.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#else
#include <memory.h>
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_buf.c	1.5	10:34:15	2/26/92";
#endif

#ifndef	MALLOC_H
extern	char	*malloc();
extern	char	*realloc();
#endif

extern	int	db_errno;

struct	head {
	int	h_fd;
	off_t	h_block;
	long	h_ticket;
	keyblk_t h_buf;
};

struct	head	*buffers;
int	nbuffers;

/*
 * isam_setbuf - set the number of ISAM block buffers
 */

int
isam_setbuf (nbuf)
int	nbuf;
{
	int	i;

	/*
	 * Check count for validity
	 */

	if (nbuf < 0)
		return (db_errno = DB_INVALID), -1;

	/*
	 * If no buffers were requested, free any existing buffers
	 */

	if (nbuf == 0) {
		if (buffers)
			free (buffers);

		buffers = 0;
		nbuffers = 0;
		return 0;
	}

	/*
	 * If no change in the number of requested buffers, do nothing
	 */

	if (nbuffers == nbuf)
		return 0;

	/*
	 * If we are adding buffers, allocate space for them and mark
	 * them as empty.
	 */

	if (nbuffers < nbuf) {
		if (nbuffers == 0)
			buffers = (struct head *)
				malloc (nbuf * sizeof (struct head));
		else
			buffers = (struct head *) realloc (buffers,
				nbuf * sizeof (struct head));

		if (buffers == 0)
			return (db_errno = DB_NOMEM), -1;

		for (i = nbuffers;i < nbuf;i++)
			buffers[i].h_fd = -1;

		nbuffers = nbuf;
	} else {
		if (! (buffers = (struct head *) realloc (buffers,
				nbuf * sizeof (struct head))))
			return (db_errno = DB_NOMEM), -1;

		nbuffers = nbuf;
	}
	return 0;
}

/*
 * __isam_getbuf - get ISAM block data associated with file
 */

int
__isam_getbuf (file, block, data)
file_t	*file;
off_t	block;
keyblk_t *data;
{
	int	hash;

	if (nbuffers == 0 || buffers == 0)
		return -1;

	hash = (file->f_key_fd ^ block) % nbuffers;
	if (buffers[hash].h_fd != file->f_key_fd ||
			buffers[hash].h_block != block ||
			buffers[hash].h_ticket != file->f_ticket)
		return -1;

	memcpy (data, &buffers[hash].h_buf, file->f_bsize);
	return 0;
}

/*
 * __isam_putbuf - save ISAM block data associated with file
 */

int
__isam_putbuf (file, block, data)
file_t	*file;
off_t	block;
keyblk_t *data;
{
	int	hash;

	if (nbuffers == 0 || buffers == 0)
		return -1;

	hash = (file->f_key_fd ^ block) % nbuffers;
	buffers[hash].h_fd = file->f_key_fd;
	buffers[hash].h_block = block;
	buffers[hash].h_ticket = file->f_ticket;
	memcpy (&buffers[hash].h_buf, data, file->f_bsize);
	return 0;
}

/*
 * __isam_endbuf - invalidate ISAM block data associated with file
 */

int
__isam_endbuf (file)
file_t	*file;
{
	int	i;

	for (i = 0;i < nbuffers;i++) {
		if (buffers[i].h_fd == file->f_key_fd)
			buffers[i].h_fd = -1;
	}
	return 0;
}
